import aicube
import os
import ujson
from time import *
import nncase_runtime as nn
import ulab.numpy as np
import time
import image
import gc
import utime

root_path="/sdcard/mp_deployment_source/"      #root_path要以/结尾
ocrdet_config_path=root_path+"ocrdet_deploy_config.json"
ocrrec_config_path=root_path+"ocrrec_deploy_config.json"
dict_path=root_path+"dict.txt"
image_path=root_path+"test.jpg"
deploy_conf={}
debug_mode=1

class ScopedTiming:
    def __init__(self, info="", enable_profile=True):
        self.info = info
        self.enable_profile = enable_profile

    def __enter__(self):
        if self.enable_profile:
            self.start_time = time.time_ns()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self.enable_profile:
            elapsed_time = time.time_ns() - self.start_time
            print(f"{self.info} took {elapsed_time / 1000000:.2f} ms")

def read_img(img_path):
    img_data = image.Image(img_path)
    img_data_rgb888=img_data.to_rgb888()
    img_hwc=img_data_rgb888.to_numpy_ref()
    shape=img_hwc.shape
    img_tmp = img_hwc.reshape((shape[0] * shape[1], shape[2]))
    img_tmp_trans = img_tmp.transpose()
    img_res=img_tmp_trans.copy()
    img_return=img_res.reshape((1,shape[2],shape[0],shape[1]))
    return img_return

def get_pad_one_side_param(out_img_size,input_img_size):
    # 右padding或下padding
    dst_w = out_img_size[0]
    dst_h = out_img_size[1]

    input_width = input_img_size[0]
    input_high = input_img_size[1]

    ratio_w = dst_w / input_width
    ratio_h = dst_h / input_high
    if ratio_w < ratio_h:
        ratio = ratio_w
    else:
        ratio = ratio_h

    new_w = (int)(ratio * input_width)
    new_h = (int)(ratio * input_high)
    dw = (dst_w - new_w) / 2
    dh = (dst_h - new_h) / 2

    top = (int)(round(0))
    bottom = (int)(round(dh * 2 + 0.1))
    left = (int)(round(0))
    right = (int)(round(dw * 2 - 0.1))
    return [0, 0, 0, 0, top, bottom, left, right]



# 读取deploy_config.json文件
def read_deploy_config(config_path):
    # 打开JSON文件以进行读取deploy_config
    with open(config_path, 'r') as json_file:
        try:
            # 从文件中加载JSON数据
            config = ujson.load(json_file)

            # 打印数据（可根据需要执行其他操作）
            #print(config)
        except ValueError as e:
            print("JSON 解析错误:", e)
    return config

def ocr():
    print("--------------start-----------------")
    deploy_conf = read_deploy_config(ocrdet_config_path)
    kmodel_name = deploy_conf["kmodel_path"]
    mask_threshold = deploy_conf["mask_threshold"]
    box_threshold = deploy_conf["box_threshold"]
    img_size = deploy_conf["img_size"]
    deploy_conf_rec = read_deploy_config(ocrrec_config_path)
    kmodel_name_rec = deploy_conf_rec["kmodel_path"]
    dict_rec = dict_path
    img_size_rec = deploy_conf_rec["img_size"]
    DICT = {}
    with open(dict_rec, 'r') as file:
        line_one = file.read(100000)
        line_list = line_one.split("\n")
    DICT = {num: char.replace("\r", "").replace("\n", "") for num, char in enumerate(line_list)}


    # ai2d输入输出初始化
    ai2d_input = read_img(image_path)
    ai2d_input_tensor = nn.from_numpy(ai2d_input)
    ai2d_input_shape=ai2d_input.shape
    data = np.ones((1,3,img_size[0],img_size[1]),dtype=np.uint8)
    ai2d_out_tensor = nn.from_numpy(data)

    # init kpu and load kmodel
    kpu_rec = nn.kpu()
    kpu_rec.load_kmodel(root_path+kmodel_name_rec)
    kpu = nn.kpu()
    ai2d = nn.ai2d()
    kpu.load_kmodel(root_path + kmodel_name)
    ai2d.set_dtype(nn.ai2d_format.NCHW_FMT,
                   nn.ai2d_format.NCHW_FMT,
                   np.uint8, np.uint8)
    ai2d.set_pad_param(True, get_pad_one_side_param(img_size,[ai2d_input_shape[3],ai2d_input_shape[2]]), 0, [0, 0, 0])
    ai2d.set_resize_param(True, nn.interp_method.tf_bilinear, nn.interp_mode.half_pixel)
    ai2d_builder = ai2d.build([1, 3, ai2d_input_shape[2], ai2d_input_shape[3]], [1, 3, img_size[0], img_size[1]])

    ai2d_rec = nn.ai2d()
    ai2d_rec.set_dtype(nn.ai2d_format.RGB_packed,
                       nn.ai2d_format.NCHW_FMT,
                       np.uint8, np.uint8)

    ai2d_input_tensor_rec = None
    ai2d_out_data_rec = np.ones((1, 3, img_size_rec[1], img_size_rec[0]), dtype=np.uint8)
    ai2d_out_tensor_rec = nn.from_numpy(ai2d_out_data_rec)
    ai2d_rec_builder = None
    with ScopedTiming("total",debug_mode > 0):
        ai2d_builder.run(ai2d_input_tensor, ai2d_out_tensor)
        kpu.set_input_tensor(0, ai2d_out_tensor)
        kpu.run()
        del ai2d_input_tensor
        del ai2d_out_tensor
        # 获取分类结果
        results = []
        for i in range(kpu.outputs_size()):
            data = kpu.get_output_tensor(i)
            result = data.to_numpy()
            del data
            results.append(result)

        tmp = (ai2d_input.shape[1], ai2d_input.shape[2], ai2d_input.shape[3])
        ai2d_input = ai2d_input.reshape((ai2d_input.shape[1], ai2d_input.shape[2] * ai2d_input.shape[3]))
        ai2d_input = ai2d_input.transpose()
        tmp2 = ai2d_input.copy()
        tmp2 = tmp2.reshape((tmp[1], tmp[2], tmp[0]))
        output_data = results[0][:, :, :, 0]
        del ai2d_input
        # 调用aicube.ocr_post_process接口进行后处理
        mp_list = aicube.ocr_post_process(output_data.reshape(-1), tmp2.reshape(-1),
                                          [img_size[0], img_size[1]],
                                          [ai2d_input_shape[3], ai2d_input_shape[2]], mask_threshold, box_threshold)

        for result_det in mp_list:
            ai2d_rec.set_pad_param(True,
                                   get_pad_one_side_param(img_size_rec, [result_det[0].shape[2], result_det[0].shape[1]]),
                                   0, [0, 0, 0])
            ai2d_rec.set_resize_param(True, nn.interp_method.tf_bilinear, nn.interp_mode.half_pixel)
            ai2d_rec_builder = ai2d_rec.build(
                [result_det[0].shape[0], result_det[0].shape[1], result_det[0].shape[2], result_det[0].shape[3]],
                [1, 3, img_size_rec[1], img_size_rec[0]])
            ai2d_input_tensor_rec = nn.from_numpy(result_det[0])
            ai2d_rec_builder.run(ai2d_input_tensor_rec, ai2d_out_tensor_rec)

            kpu_rec.set_input_tensor(0, ai2d_out_tensor_rec)
            kpu_rec.run()
            results = []
            for i in range(kpu_rec.outputs_size()):
                data = kpu_rec.get_output_tensor(i)
                result = data.to_numpy()
                del data
                results.append(result)
            preds = np.argmax(results[0], axis=2).reshape((-1))
            output_txt = ""
            for i in range(len(preds)):
                if preds[i] != (len(DICT) - 1) and (not (i > 0 and preds[i - 1] == preds[i])):
                    output_txt = output_txt + DICT[preds[i]]
            print("\n" + output_txt + "\n")
            print("-------------------------------")

        image_draw = image.Image(image_path).to_rgb565()
        if mp_list:
            for j in mp_list:
                for i in range(4):
                    x1 = j[1][(i * 2)]
                    y1 = j[1][(i * 2 + 1)]
                    x2 = j[1][((i + 1) * 2) % 8]
                    y2 = j[1][((i + 1) * 2 + 1) % 8]
                    image_draw.draw_line((int(x1), int(y1), int(x2), int(y2)), color=(255, 0, 0, 255), thickness=5)
        image_draw.compress_for_ide()
        image_draw.save(root_path + "ocr_result.jpg")
        del ai2d_input_tensor_rec
        del ai2d_out_tensor_rec
    del ai2d
    del ai2d_builder
    del ai2d_rec
    if ai2d_rec_builder:
        del ai2d_rec_builder
    del kpu
    del kpu_rec
    print("---------------end------------------")
    gc.collect()


if __name__=="__main__":
    ocr()
