import ulab.numpy as np
import nncase_runtime as nn
from media.camera import *
from media.display import *
from media.media import *
import aicube
import image
import time
import gc
import ujson
import os, sys

#display分辨率
DISPLAY_WIDTH = ALIGN_UP(1920, 16)
DISPLAY_HEIGHT = 1080

#ai原图分辨率输入
OUT_RGB888P_WIDTH = ALIGN_UP(512, 16)
OUT_RGB888P_HEIGH = 512

root_path= "/sdcard/mp_deployment_source/" #root_path要以/结尾
config_path= root_path+"deploy_config.json"
deploy_conf={}
debug_mode=1

class ScopedTiming:
    def __init__(self, info="", enable_profile=True):
        self.info = info
        self.enable_profile = enable_profile

    def __enter__(self):
        if self.enable_profile:
            self.start_time = time.time_ns()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self.enable_profile:
            elapsed_time = time.time_ns() - self.start_time
            print(f"{self.info} took {elapsed_time / 1000000:.2f} ms")

def read_deploy_config(config_path):
    # 打开JSON文件以进行读取deploy_config
    with open(config_path, 'r') as json_file:
        try:
            # 从文件中加载JSON数据
            config = ujson.load(json_file)

            # 打印数据（可根据需要执行其他操作）
            #print(config)
        except ValueError as e:
            print("JSON 解析错误:", e)
    return config

def segmentation():
    print("start")

    # 使用json读取内容初始化部署变量
    deploy_conf=read_deploy_config(config_path)
    kmodel_name=deploy_conf["kmodel_path"]
    labels=deploy_conf["categories"]
    confidence_threshold=deploy_conf["confidence_threshold"]
    img_size=deploy_conf["img_size"]
    num_classes=deploy_conf["num_classes"]
    cls_idx=-1

    # init kpu and load kmodel
    kpu = nn.kpu()

    kpu.load_kmodel(root_path+kmodel_name)
    # use hdmi for display
    display.init(LT9611_1920X1080_30FPS)

    config = k_vb_config()
    config.max_pool_cnt = 1
    config.comm_pool[0].blk_size = 4*DISPLAY_WIDTH*DISPLAY_HEIGHT
    config.comm_pool[0].blk_cnt = 1
    config.comm_pool[0].mode = VB_REMAP_MODE_NOCACHE

    media.buffer_config(config)

    camera.sensor_init(CAM_DEV_ID_0, CAM_DEFAULT_SENSOR)

    # set chn0 output yuv420sp
    camera.set_outsize(CAM_DEV_ID_0, CAM_CHN_ID_0, DISPLAY_WIDTH, DISPLAY_HEIGHT)
    camera.set_outfmt(CAM_DEV_ID_0, CAM_CHN_ID_0, PIXEL_FORMAT_YUV_SEMIPLANAR_420)

    meida_source = media_device(CAMERA_MOD_ID, CAM_DEV_ID_0, CAM_CHN_ID_0)
    meida_sink = media_device(DISPLAY_MOD_ID, DISPLAY_DEV_ID, DISPLAY_CHN_VIDEO1)
    media.create_link(meida_source, meida_sink)

    display.set_plane(0, 0, DISPLAY_WIDTH, DISPLAY_HEIGHT, PIXEL_FORMAT_YVU_PLANAR_420, DISPLAY_MIRROR_NONE, DISPLAY_CHN_VIDEO1)

    # set chn2 output rgb88planar
    camera.set_outsize(CAM_DEV_ID_0, CAM_CHN_ID_2, OUT_RGB888P_WIDTH, OUT_RGB888P_HEIGH)
    camera.set_outfmt(CAM_DEV_ID_0, CAM_CHN_ID_2, PIXEL_FORMAT_RGB_888_PLANAR)

    try:
        media.buffer_init()

        buffer = media.request_buffer(4*DISPLAY_WIDTH*DISPLAY_HEIGHT)
        draw_img = image.Image(DISPLAY_WIDTH, DISPLAY_HEIGHT, image.ARGB8888)
        osd_img = image.Image(DISPLAY_WIDTH, DISPLAY_HEIGHT, image.ARGB8888, poolid=buffer.pool_id, alloc=image.ALLOC_VB, phyaddr=buffer.phys_addr, virtaddr=buffer.virt_addr)
        camera.start_stream(CAM_DEV_ID_0)
        ai2d_input_tensor=None
        while  True:
            os.exitpoint()
            with ScopedTiming("total",debug_mode > 0):
                rgb888p_img = camera.capture_image(CAM_DEV_ID_0, CAM_CHN_ID_2)

                # for rgb888planar
                if rgb888p_img.format() == image.RGBP888:
                    ai2d_input = rgb888p_img.to_numpy_ref()
                    ai2d_input_tensor = nn.from_numpy(ai2d_input)
                    # set input
                    kpu.set_input_tensor(0, ai2d_input_tensor)
                    # run kmodel
                    kpu.run()
                    # get output
                    results = []
                    for i in range(kpu.outputs_size()):
                        out_data = kpu.get_output_tensor(i)
                        result = out_data.to_numpy()[0]
                        del out_data
                        results.append(result)
                    output = results[0]
                    img = aicube.seg_post_process(output, num_classes, [OUT_RGB888P_HEIGH, OUT_RGB888P_WIDTH], [DISPLAY_HEIGHT,DISPLAY_WIDTH])
                draw_img = image.Image(DISPLAY_WIDTH, DISPLAY_HEIGHT, image.ARGB8888,alloc=image.ALLOC_REF,data=img) #result
                draw_img.copy_to(osd_img)
                display.show_image(osd_img, 0, 0, DISPLAY_CHN_OSD3)
                camera.release_image(CAM_DEV_ID_0, CAM_CHN_ID_2, rgb888p_img)
                gc.collect()
    except KeyboardInterrupt as e:
        print("user stop: ", e)
    except BaseException as e:
        sys.print_exception(e)
    finally:
        camera.stop_stream(CAM_DEV_ID_0)
        display.deinit()

        if ai2d_input_tensor:
            del ai2d_input_tensor
        del kpu
        gc.collect()
        os.exitpoint(os.EXITPOINT_ENABLE_SLEEP)
        time.sleep_ms(100)
        media.release_buffer(buffer)
        media.destroy_link(meida_source, meida_sink)
        media.buffer_deinit()
    print("end")
    return 0


def softmax(x,axis):
    x -= np.max(x, axis = axis)
    x = np.exp(x) / np.sum(np.exp(x), axis = axis)
    return x

def seg_postprocess(x,img_size):
    x_softmax = softmax(x, axis=1)
    result = np.argmax(x_softmax, axis=1)
    result_shape = np.zeros((1,1,img_size[0],img_size[1]),dtype=np.uint8)
    result_shape[0] = result
    return result_shape

if __name__ == "__main__":
    os.exitpoint(os.EXITPOINT_ENABLE)
    segmentation()

